@@ -19,7 +19,7 @@
  * Boston, MA 02111-1307, USA.
  */
 
-#include <config.h>
+#include "config.h"
 
 #include "pango-impl-utils.h"
 #include "pango-glyph.h"
@@ -50,7 +50,7 @@ pango_shape (const gchar      *text,
 
   glyphs->num_glyphs = 0;
 
-  if (G_LIKELY (PANGO_IS_ENGINE_SHAPE (analysis->shape_engine) && PANGO_IS_FONT (analysis->font)))
+  if (G_LIKELY (analysis->shape_engine && analysis->font))
     {
       _pango_engine_shape_shape (analysis->shape_engine, analysis->font,
 				 text, length, analysis, glyphs);
@@ -106,37 +106,23 @@ pango_shape (const gchar      *text,
 	}
     }
   else
-    {
-      if (!PANGO_IS_ENGINE_SHAPE (analysis->shape_engine) &&
-	  !_pango_warning_history.shape_shape_engine)
-	{
-	  _pango_warning_history.shape_shape_engine = TRUE;
-	  g_warning ("pango_shape called with bad shape_engine, expect ugly output");
-	}
-      if (!PANGO_IS_FONT (analysis->font) &&
-	  !_pango_warning_history.shape_font)
-	{
-	  _pango_warning_history.shape_font = TRUE;
-	  g_warning ("pango_shape called with bad font, expect ugly output");
-	}
+    glyphs->num_glyphs = 0;
 
-      glyphs->num_glyphs = 0;
-    }
-
-  if (!glyphs->num_glyphs)
+  if (G_UNLIKELY (!glyphs->num_glyphs))
     {
       PangoEngineShape *fallback_engine = _pango_get_fallback_shaper ();
 
       _pango_engine_shape_shape (fallback_engine, analysis->font,
 				 text, length, analysis, glyphs);
+      if (G_UNLIKELY (!glyphs->num_glyphs))
+        return;
     }
 
   /* make sure last_cluster is invalid */
   last_cluster = glyphs->log_clusters[0] - 1;
   for (i = 0; i < glyphs->num_glyphs; i++)
     {
-     /* Set glyphs[i].attr.is_cluster_start based on log_clusters[]
-      */
+      /* Set glyphs[i].attr.is_cluster_start based on log_clusters[] */
       if (glyphs->log_clusters[i] != last_cluster)
 	{
 	  glyphs->glyphs[i].attr.is_cluster_start = TRUE;
@@ -156,4 +142,31 @@ pango_shape (const gchar      *text,
 	  glyphs->glyphs[i].geometry.x_offset += glyphs->glyphs[i].geometry.width;
 	}
     }
+
+  /* Make sure glyphstring direction conforms to analysis->level */
+  if (G_UNLIKELY ((analysis->level & 1) &&
+		  glyphs->log_clusters[0] < glyphs->log_clusters[glyphs->num_glyphs - 1]))
+    {
+      /* Warn once per shaper */
+      static GQuark warned_quark = 0;
+
+      if (!warned_quark)
+	warned_quark = g_quark_from_static_string ("pango-shape-warned");
+
+      if (analysis->shape_engine && !g_object_get_qdata (G_OBJECT (analysis->shape_engine), warned_quark))
+	{
+	  GType engine_type = G_OBJECT_TYPE (analysis->shape_engine);
+	  const char *engine_name = g_type_name (engine_type);
+	  if (!engine_name)
+	    engine_name = "(unknown)";
+
+	  g_warning ("Expected RTL run but shape-engine='%s' returned LTR. Fixing.", engine_name);
+
+	  g_object_set_qdata_full (G_OBJECT (analysis->shape_engine), warned_quark,
+				   GINT_TO_POINTER (1), NULL);
+	}
+
+      /* *Fix* it so we don't crash later */
+      pango_glyph_string_reverse_range (glyphs, 0, glyphs->num_glyphs);
+    }
 }
